#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Dict, List, Mapping

from airbyte_cdk.sources.declarative.schema.schema_loader import SchemaLoader


@dataclass
class CompositeSchemaLoader(SchemaLoader):
    """
    Schema loader that consists of multiple schema loaders that are combined into a single
    schema. Subsequent schemas do not overwrite existing values so the schema loaders with
    a higher priority should be defined first.
    """

    schema_loaders: List[SchemaLoader]
    parameters: InitVar[Mapping[str, Any]]

    def get_json_schema(self) -> Mapping[str, Any]:
        combined_schema: Dict[str, Any] = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": ["null", "object"],
            "properties": {},
        }
        for schema_loader in self.schema_loaders:
            schema_properties = schema_loader.get_json_schema()["properties"]
            combined_schema["properties"] = {**schema_properties, **combined_schema["properties"]}
        return combined_schema



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import logging
from math import log
from typing import Any, ClassVar, Dict, Iterator, List, Mapping, Optional, Union

from airbyte_cdk.connector_builder.models import (
    AuxiliaryRequest,
    LogMessage,
    StreamRead,
    StreamReadSlices,
)
from airbyte_cdk.entrypoint import AirbyteEntrypoint
from airbyte_cdk.models import (
    AirbyteControlMessage,
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteStateMessage,
    AirbyteTraceMessage,
    ConfiguredAirbyteCatalog,
    TraceType,
)
from airbyte_cdk.sources.declarative.declarative_source import DeclarativeSource
from airbyte_cdk.utils import AirbyteTracedException
from airbyte_cdk.utils.datetime_format_inferrer import DatetimeFormatInferrer
from airbyte_cdk.utils.schema_inferrer import (
    SchemaInferrer,
    SchemaValidationException,
)

from .helpers import clean_config
from .message_grouper import get_message_groups
from .types import GROUPED_MESSAGES, INFERRED_SCHEMA_OUTPUT_TYPE, MESSAGE_GROUPS


class TestReader:
    """
    A utility class for performing test reads from a declarative data source, primarily used to validate
    connector configurations by performing partial stream reads.

    Initialization:

        TestReader(max_pages_per_slice: int, max_slices: int, max_record_limit: int = 1000)
            Initializes a new instance of the TestReader class with limits on pages per slice, slices, and records
            per read operation.

    Public Methods:
        run_test_read(source, config, configured_catalog, state, record_limit=None) -> StreamRead:

            Executes a test read operation from the given declarative source. It configures and infers the schema,
            processes the read messages (including logging and error handling), and returns a StreamRead object
            that contains slices of data, log messages, auxiliary requests, and any inferred schema or datetime formats.

            Parameters:
                source (DeclarativeSource): The data source to read from.
                config (Mapping[str, Any]): Configuration parameters for the source.
                configured_catalog (ConfiguredAirbyteCatalog): Catalog containing stream configuration.
                state (List[AirbyteStateMessage]): Current state information for the read.
                record_limit (Optional[int]): Optional override for the maximum number of records to read.

            Returns:
                StreamRead: An object encapsulating logs, data slices, auxiliary requests, and inferred metadata,
                along with indicators if any configured limit was reached.

    """

    __test__: ClassVar[bool] = False  # Tell Pytest this is not a Pytest class, despite its name

    logger = logging.getLogger("airbyte.connector-builder")

    def __init__(
        self,
        max_pages_per_slice: int,
        max_slices: int,
        max_record_limit: int = 1000,
    ) -> None:
        self._max_pages_per_slice = max_pages_per_slice
        self._max_slices = max_slices
        self._max_record_limit = max_record_limit

    def run_test_read(
        self,
        source: DeclarativeSource,
        config: Mapping[str, Any],
        configured_catalog: ConfiguredAirbyteCatalog,
        state: List[AirbyteStateMessage],
        record_limit: Optional[int] = None,
    ) -> StreamRead:
        """
        Run a test read for the connector by reading from a single stream and inferring schema and datetime formats.

        Parameters:
            source (DeclarativeSource): The source instance providing the streams.
            config (Mapping[str, Any]): The configuration settings to use for reading.
            configured_catalog (ConfiguredAirbyteCatalog): The catalog specifying the stream configuration.
            state (List[AirbyteStateMessage]): A list of state messages to resume the read.
            record_limit (Optional[int], optional): Maximum number of records to read. Defaults to None.

        Returns:
            StreamRead: An object containing the following attributes:
                - logs (List[str]): Log messages generated during the process.
                - slices (List[Any]): The data slices read from the stream.
                - test_read_limit_reached (bool): Indicates whether the record limit was reached.
                - auxiliary_requests (Any): Any auxiliary requests generated during reading.
                - inferred_schema (Any): The schema inferred from the stream data.
                - latest_config_update (Any): The latest configuration update, if applicable.
                - inferred_datetime_formats (Dict[str, str]): Mapping of fields to their inferred datetime formats.
        """

        record_limit = self._check_record_limit(record_limit)
        # The connector builder currently only supports reading from a single stream at a time
        stream = source.streams(config)[0]

        # get any deprecation warnings during the component creation
        deprecation_warnings: List[LogMessage] = source.deprecation_warnings()

        schema_inferrer = SchemaInferrer(
            self._pk_to_nested_and_composite_field(stream.primary_key),
            self._cursor_field_to_nested_and_composite_field(stream.cursor_field),
        )
        datetime_format_inferrer = DatetimeFormatInferrer()

        message_group = get_message_groups(
            self._read_stream(source, config, configured_catalog, state),
            schema_inferrer,
            datetime_format_inferrer,
            record_limit,
        )

        slices, log_messages, auxiliary_requests, latest_config_update = self._categorise_groups(
            message_group
        )

        # extend log messages with deprecation warnings
        log_messages.extend(deprecation_warnings)

        schema, log_messages = self._get_infered_schema(
            configured_catalog, schema_inferrer, log_messages
        )

        return StreamRead(
            logs=log_messages,
            slices=slices,
            test_read_limit_reached=self._has_reached_limit(slices),
            auxiliary_requests=auxiliary_requests,
            inferred_schema=schema,
            latest_config_update=self._get_latest_config_update(latest_config_update),
            inferred_datetime_formats=datetime_format_inferrer.get_inferred_datetime_formats(),
        )

    def _pk_to_nested_and_composite_field(
        self, field: Optional[Union[str, List[str], List[List[str]]]]
    ) -> List[List[str]]:
        """
        Converts a primary key definition into a nested list representation.

        The function accepts a primary key that can be a single string, a list of strings, or a list of lists of strings.
        It ensures that the return value is always a list of lists of strings.

        Parameters:
            field (Optional[Union[str, List[str], List[List[str]]]]):
                The primary key definition. This can be:
                  - None or an empty value: returns a list containing an empty list.
                  - A single string: returns a list containing one list with the string.
                  - A list of strings (composite key): returns a list where each key is encapsulated in its own list.
                  - A list of lists of strings (nested field structure): returns as is.

        Returns:
            List[List[str]]:
                A nested list representation of the primary key.
        """
        if not field:
            return [[]]

        if isinstance(field, str):
            return [[field]]

        is_composite_key = isinstance(field[0], str)
        if is_composite_key:
            return [[i] for i in field]  # type: ignore  # the type of field is expected to be List[str] here

        return field  # type: ignore  # the type of field is expected to be List[List[str]] here

    def _cursor_field_to_nested_and_composite_field(
        self, field: Union[str, List[str]]
    ) -> List[List[str]]:
        """
        Transforms the cursor field input into a nested list format suitable for further processing.

        This function accepts a cursor field specification, which can be either:
            - A falsy value (e.g., None or an empty string), in which case it returns a list containing an empty list.
            - A string representing a simple cursor field. The string is wrapped in a nested list.
            - A list of strings representing a composite or nested cursor field. The list is returned wrapped in an outer list.

        Parameters:
            field (Union[str, List[str]]): The cursor field specification. It can be:
                - An empty or falsy value: returns [[]].
                - A string: returns [[field]].
                - A list of strings: returns [field] if the first element is a string.

        Returns:
            List[List[str]]: A nested list representation of the cursor field.

        Raises:
            ValueError: If the input is a list but its first element is not a string,
                        indicating an unsupported type for a cursor field.
        """
        if not field:
            return [[]]

        if isinstance(field, str):
            return [[field]]

        is_nested_key = isinstance(field[0], str)
        if is_nested_key:
            return [field]

        raise ValueError(f"Unknown type for cursor field `{field}")

    def _check_record_limit(self, record_limit: Optional[int] = None) -> int:
        """
        Checks and adjusts the provided record limit to ensure it falls within the valid range.

        If record_limit is provided, it must be between 1 and self._max_record_limit inclusive.
        If record_limit is None, it defaults to self._max_record_limit.

        Args:
            record_limit (Optional[int]): The requested record limit to validate.

        Returns:
            int: The validated record limit. If record_limit exceeds self._max_record_limit, the maximum allowed value is used.

        Raises:
            ValueError: If record_limit is provided and is not between 1 and self._max_record_limit.
        """
        if record_limit is not None and not (1 <= record_limit <= self._max_record_limit):
            raise ValueError(
                f"Record limit must be between 1 and {self._max_record_limit}. Got {record_limit}"
            )

        if record_limit is None:
            record_limit = self._max_record_limit
        else:
            record_limit = min(record_limit, self._max_record_limit)

        return record_limit

    def _categorise_groups(self, message_groups: MESSAGE_GROUPS) -> GROUPED_MESSAGES:
        """
        Categorizes a sequence of message groups into slices, log messages, auxiliary requests, and the latest configuration update.

        This function iterates over each message group in the provided collection and processes it based on its type:
            - AirbyteLogMessage: Converts the log message into a LogMessage object and appends it to the log_messages list.
            - AirbyteTraceMessage with type ERROR: Extracts error details, creates a LogMessage at the "ERROR" level, and appends it.
            - AirbyteControlMessage: Updates the latest_config_update if the current message is more recent.
            - AuxiliaryRequest: Appends the message to the auxiliary_requests list.
            - StreamReadSlices: Appends the message to the slices list.
            - Any other type: Raises a ValueError indicating an unknown message group type.

        Parameters:
            message_groups (MESSAGE_GROUPS): A collection of message groups to be processed.

        Returns:
            GROUPED_MESSAGES: A tuple containing four elements:
                - slices: A list of StreamReadSlices objects.
                - log_messages: A list of LogMessage objects.
                - auxiliary_requests: A list of AuxiliaryRequest objects.
                - latest_config_update: The most recent AirbyteControlMessage, or None if none was processed.

        Raises:
            ValueError: If a message group of an unknown type is encountered.
        """

        slices = []
        log_messages = []
        auxiliary_requests = []
        latest_config_update: Optional[AirbyteControlMessage] = None

        # process the message groups first
        for message_group in message_groups:
            match message_group:
                case AirbyteLogMessage():
                    log_messages.append(
                        LogMessage(message=message_group.message, level=message_group.level.value)
                    )
                case AirbyteTraceMessage():
                    if message_group.type == TraceType.ERROR:
                        log_messages.append(
                            LogMessage(
                                message=message_group.error.message,  # type: ignore
                                level="ERROR",
                                internal_message=message_group.error.internal_message,  # type: ignore
                                stacktrace=message_group.error.stack_trace,  # type: ignore
                            )
                        )
                case AirbyteControlMessage():
                    if (
                        not latest_config_update
                        or latest_config_update.emitted_at <= message_group.emitted_at
                    ):
                        latest_config_update = message_group
                case AuxiliaryRequest():
                    auxiliary_requests.append(message_group)
                case StreamReadSlices():
                    slices.append(message_group)
                case _:
                    raise ValueError(f"Unknown message group type: {type(message_group)}")

        return slices, log_messages, auxiliary_requests, latest_config_update

    def _get_infered_schema(
        self,
        configured_catalog: ConfiguredAirbyteCatalog,
        schema_inferrer: SchemaInferrer,
        log_messages: List[LogMessage],
    ) -> INFERRED_SCHEMA_OUTPUT_TYPE:
        """
        Retrieves the inferred schema from the given configured catalog using the provided schema inferrer.

        This function processes a single stream from the configured catalog. It attempts to obtain the stream's
        schema via the schema inferrer. If a SchemaValidationException occurs, each validation error is logged in the
        provided log_messages list and the partially inferred schema (from the exception) is returned.

        Parameters:
            configured_catalog (ConfiguredAirbyteCatalog): The configured catalog that contains the stream metadata.
                It is assumed that only one stream is defined.
            schema_inferrer (SchemaInferrer): An instance responsible for inferring the schema for a given stream.
            log_messages (List[LogMessage]): A list that will be appended with log messages, especially error messages
                if schema validation issues arise.

        Returns:
            INFERRED_SCHEMA_OUTPUT_TYPE: A tuple consisting of the inferred schema and the updated list of log messages.
        """

        try:
            # The connector builder currently only supports reading from a single stream at a time
            configured_stream = configured_catalog.streams[0]
            schema = schema_inferrer.get_stream_schema(configured_stream.stream.name)
        except SchemaValidationException as exception:
            # we update the log_messages with possible errors
            for validation_error in exception.validation_errors:
                log_messages.append(LogMessage(validation_error, "ERROR"))
            schema = exception.schema

        return schema, log_messages

    def _get_latest_config_update(
        self,
        latest_config_update: AirbyteControlMessage | None,
    ) -> Dict[str, Any] | None:
        """
        Retrieves a cleaned configuration from the latest Airbyte control message.

        This helper function extracts the configuration from the given Airbyte control message, cleans it using the internal `Parsers.clean_config` function,
        and returns the resulting dictionary. If no control message is provided (i.e., latest_config_update is None), the function returns None.

        Parameters:
            latest_config_update (AirbyteControlMessage | None): The control message containing the connector configuration. May be None.

        Returns:
            Dict[str, Any] | None: The cleaned configuration dictionary if available; otherwise, None.
        """

        return (
            clean_config(latest_config_update.connectorConfig.config)  # type: ignore
            if latest_config_update
            else None
        )

    def _read_stream(
        self,
        source: DeclarativeSource,
        config: Mapping[str, Any],
        configured_catalog: ConfiguredAirbyteCatalog,
        state: List[AirbyteStateMessage],
    ) -> Iterator[AirbyteMessage]:
        """
        Reads messages from the given DeclarativeSource using an AirbyteEntrypoint.

        This method attempts to yield messages from the source's read generator. If the generator
        raises an AirbyteTracedException, it checks whether the exception message indicates a non-actionable
        error (e.g., a final exception from AbstractSource that should not be logged). In that case, it stops
        processing without yielding the exception as a message. For other exceptions, the exception is caught,
        wrapped into an AirbyteTracedException, and yielded as an AirbyteMessage.

        Parameters:
            source (DeclarativeSource): The source object that provides data reading logic.
            config (Mapping[str, Any]): The configuration dictionary for the source.
            configured_catalog (ConfiguredAirbyteCatalog): The catalog defining the streams and their configurations.
            state (List[AirbyteStateMessage]): A list representing the current state for incremental sync.

        Yields:
            AirbyteMessage: Messages yielded from the source's generator. In case of exceptions,
            an AirbyteMessage encapsulating the error is yielded instead.
        """
        # the generator can raise an exception
        # iterate over the generated messages. if next raise an exception, catch it and yield it as an AirbyteLogMessage
        try:
            yield from AirbyteEntrypoint(source).read(
                source.spec(self.logger), config, configured_catalog, state
            )
        except AirbyteTracedException as traced_exception:
            # Look for this message which indicates that it is the "final exception" raised by AbstractSource.
            # If it matches, don't yield this as we don't need to show this in the Builder.
            # This is somewhat brittle as it relies on the message string, but if they drift then the worst case
            # is that this message will be shown in the Builder.
            if (
                traced_exception.message is not None
                and "During the sync, the following streams did not sync successfully"
                in traced_exception.message
            ):
                return
            yield traced_exception.as_airbyte_message()
        except Exception as e:
            error_message = f"{e.args[0] if len(e.args) > 0 else str(e)}"
            yield AirbyteTracedException.from_exception(
                e, message=error_message
            ).as_airbyte_message()

    def _has_reached_limit(self, slices: List[StreamReadSlices]) -> bool:
        """
        Determines whether the provided collection of slices has reached any defined limits.

        This function checks for three types of limits:
        1. If the number of slices is greater than or equal to a maximum slice limit.
        2. If any individual slice has a number of pages that meets or exceeds a maximum number of pages per slice.
        3. If the cumulative number of records across all pages in all slices reaches or exceeds a maximum record limit.

        Parameters:
            slices (List[StreamReadSlices]): A list where each element represents a slice containing one or more pages, and each page has a collection of records.

        Returns:
            bool: True if any of the following conditions is met:
                - The number of slices is at or above the maximum allowed slices.
                - Any slice contains pages at or above the maximum allowed per slice.
                - The total count of records reaches or exceeds the maximum record limit.
            False otherwise.
        """
        if len(slices) >= self._max_slices:
            return True

        record_count = 0

        for _slice in slices:
            if len(_slice.pages) >= self._max_pages_per_slice:
                return True
            for page in _slice.pages:
                record_count += len(page.records)
                if record_count >= self._max_record_limit:
                    return True
        return False



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Airbyte CDK 'image' commands.

The `airbyte-cdk image build` command provides a simple way to work with Airbyte
connector images.
"""

import sys
from pathlib import Path

import rich_click as click

from airbyte_cdk.models.connector_metadata import MetadataFile
from airbyte_cdk.utils.connector_paths import resolve_connector_name_and_directory
from airbyte_cdk.utils.docker import (
    ConnectorImageBuildError,
    build_connector_image,
    verify_docker_installation,
)


@click.group(
    name="image",
    help=__doc__.replace("\n", "\n\n"),  # Render docstring as help text (markdown)
)
def image_cli_group() -> None:
    """Commands for working with connector Docker images."""


@image_cli_group.command()
@click.argument(
    "connector",
    required=False,
    type=str,
    metavar="[CONNECTOR]",
)
@click.option("--tag", default="dev", help="Tag to apply to the built image (default: dev)")
@click.option("--no-verify", is_flag=True, help="Skip verification of the built image")
@click.option(
    "--dockerfile",
    type=click.Path(exists=True, file_okay=True, path_type=Path),
    help="Optional. Override the Dockerfile used for building the image.",
)
def build(
    connector: str | None = None,
    *,
    tag: str = "dev",
    no_verify: bool = False,
    dockerfile: Path | None = None,
) -> None:
    """Build a connector Docker image.

    [CONNECTOR] can be a connector name (e.g. 'source-pokeapi'), a path to a connector directory, or omitted to use the current working directory.
    If a string containing '/' is provided, it is treated as a path. Otherwise, it is treated as a connector name.
    """
    if not verify_docker_installation():
        click.echo(
            "Docker is not installed or not running. Please install Docker and try again.", err=True
        )
        sys.exit(1)

    connector_name, connector_directory = resolve_connector_name_and_directory(connector)

    metadata_file_path: Path = connector_directory / "metadata.yaml"
    try:
        metadata = MetadataFile.from_file(metadata_file_path)
    except (FileNotFoundError, ValueError) as e:
        click.echo(
            f"Error loading metadata file '{metadata_file_path}': {e!s}",
            err=True,
        )
        sys.exit(1)
    click.echo(f"Building Image for Connector: {metadata.data.dockerRepository}:{tag}")
    try:
        build_connector_image(
            connector_directory=connector_directory,
            connector_name=connector_name,
            metadata=metadata,
            tag=tag,
            no_verify=no_verify,
            dockerfile_override=dockerfile or None,
        )
    except ConnectorImageBuildError as e:
        click.echo(
            f"Error building connector image: {e!s}",
            err=True,
        )
        sys.exit(1)


__all__ = [
    "image_cli_group",
]



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from dataclasses import InitVar, dataclass
from typing import Any, Iterable, Mapping, Optional

from airbyte_cdk.sources.declarative.incremental import DeclarativeCursor
from airbyte_cdk.sources.declarative.types import Record, StreamSlice, StreamState
from airbyte_cdk.sources.streams.checkpoint.checkpoint_reader import FULL_REFRESH_COMPLETE_STATE


@dataclass
class ResumableFullRefreshCursor(DeclarativeCursor):
    parameters: InitVar[Mapping[str, Any]]

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._cursor: StreamState = {}

    def get_stream_state(self) -> StreamState:
        return self._cursor

    def set_initial_state(self, stream_state: StreamState) -> None:
        self._cursor = stream_state

    def observe(self, stream_slice: StreamSlice, record: Record) -> None:
        """
        Resumable full refresh manages state using a page number so it does not need to update state by observing incoming records.
        """
        pass

    def close_slice(self, stream_slice: StreamSlice, *args: Any) -> None:
        # The ResumableFullRefreshCursor doesn't support nested streams yet so receiving a partition is unexpected
        if stream_slice.partition:
            raise ValueError(
                f"Stream slice {stream_slice} should not have a partition. Got {stream_slice.partition}."
            )
        self._cursor = stream_slice.cursor_slice

    def should_be_synced(self, record: Record) -> bool:
        """
        Unlike date-based cursors which filter out records outside slice boundaries, resumable full refresh records exist within pages
        that don't have filterable bounds. We should always return them.
        """
        return True

    def is_greater_than_or_equal(self, first: Record, second: Record) -> bool:
        """
        RFR record don't have ordering to be compared between one another.
        """
        return False

    def select_state(self, stream_slice: Optional[StreamSlice] = None) -> Optional[StreamState]:
        # A top-level RFR cursor only manages the state of a single partition
        return self._cursor

    def stream_slices(self) -> Iterable[StreamSlice]:
        """
        Resumable full refresh cursors only return a single slice and can't perform partitioning because iteration is done per-page
        along an unbounded set.
        """
        yield from [StreamSlice(cursor_slice=self._cursor, partition={})]

    # This is an interesting pattern that might not seem obvious at first glance. This cursor itself has no functional need to
    # inject any request values into the outbound response because the up-to-date pagination state is already loaded and
    # maintained by the paginator component
    def get_request_params(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_headers(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_body_data(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_body_json(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}


@dataclass
class ChildPartitionResumableFullRefreshCursor(ResumableFullRefreshCursor):
    """
    The Sub-stream Resumable Cursor for Full-Refresh substreams.
    Follows the parent type `ResumableFullRefreshCursor` with a small override,
    to provide the ability to close the substream's slice once it has finished processing.

    Check the `close_slice` method overide for more info about the actual behaviour of this cursor.
    """

    def close_slice(self, stream_slice: StreamSlice, *args: Any) -> None:
        """
        Once the current slice has finished syncing:
         - paginator returns None
         - no more slices to process

        we assume that the records are processed and emitted already,
        thus we have to set the cursor to ` __ab_full_refresh_sync_complete: true `,
        otherwise there is a risk of Inf. Loop processing the same slice.
        """
        self._cursor = FULL_REFRESH_COMPLETE_STATE



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.partition_routers.async_job_partition_router import (
    AsyncJobPartitionRouter,
)
from airbyte_cdk.sources.declarative.partition_routers.cartesian_product_stream_slicer import (
    CartesianProductStreamSlicer,
)
from airbyte_cdk.sources.declarative.partition_routers.grouping_partition_router import (
    GroupingPartitionRouter,
)
from airbyte_cdk.sources.declarative.partition_routers.list_partition_router import (
    ListPartitionRouter,
)
from airbyte_cdk.sources.declarative.partition_routers.partition_router import PartitionRouter
from airbyte_cdk.sources.declarative.partition_routers.single_partition_router import (
    SinglePartitionRouter,
)
from airbyte_cdk.sources.declarative.partition_routers.substream_partition_router import (
    SubstreamPartitionRouter,
)

__all__ = [
    "AsyncJobPartitionRouter",
    "CartesianProductStreamSlicer",
    "GroupingPartitionRouter",
    "ListPartitionRouter",
    "SinglePartitionRouter",
    "SubstreamPartitionRouter",
    "PartitionRouter",
]



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Module for future CDK components.

Components here are planned to move to the CDK.

TODO!: Add GitHub link here before merging.
"""

from __future__ import annotations

from airbyte_cdk.sql.shared.sql_processor import SqlProcessorBase

__all__ = [
    "SqlProcessorBase",
]



"""Docker build utilities for Airbyte CDK."""

from __future__ import annotations

import json
import logging
import os
import subprocess
import sys
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import click
import requests

from airbyte_cdk.models.connector_metadata import ConnectorLanguage, MetadataFile
from airbyte_cdk.utils.connector_paths import resolve_airbyte_repo_root


@dataclass(kw_only=True)
class ConnectorImageBuildError(Exception):
    """Custom exception for Docker build errors."""

    error_text: str
    build_args: list[str]

    def __str__(self) -> str:
        return "\n".join(
            [
                f"ConnectorImageBuildError: Could not build image.",
                f"Build args: {self.build_args}",
                f"Error text: {self.error_text}",
            ]
        )


logger = logging.getLogger(__name__)


class ArchEnum(str, Enum):
    """Enum for supported architectures."""

    ARM64 = "arm64"
    AMD64 = "amd64"


def _build_image(
    context_dir: Path,
    dockerfile: Path,
    metadata: MetadataFile,
    tag: str,
    arch: ArchEnum,
    build_args: dict[str, str | None] | None = None,
) -> str:
    """Build a Docker image for the specified architecture.

    Returns the tag of the built image.

    Raises: ConnectorImageBuildError if the build fails.
    """
    docker_args: list[str] = [
        "docker",
        "build",
        "--platform",
        f"linux/{arch.value}",
        "--file",
        str(dockerfile),
        "--label",
        f"io.airbyte.version={metadata.data.dockerImageTag}",
        "--label",
        f"io.airbyte.name={metadata.data.dockerRepository}",
    ]
    if build_args:
        for key, value in build_args.items():
            if value is not None:
                docker_args.append(f"--build-arg={key}={value}")
            else:
                docker_args.append(f"--build-arg={key}")
    docker_args.extend(
        [
            "-t",
            tag,
            str(context_dir),
        ]
    )

    print(f"Building image: {tag} ({arch})")
    try:
        run_docker_command(
            docker_args,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise ConnectorImageBuildError(
            error_text=e.stderr,
            build_args=docker_args,
        ) from e

    return tag


def _tag_image(
    tag: str,
    new_tags: list[str] | str,
) -> None:
    """Build a Docker image for the specified architecture.

    Returns the tag of the built image.

    Raises:
        ConnectorImageBuildError: If the docker tag command fails.
    """
    if not isinstance(new_tags, list):
        new_tags = [new_tags]

    for new_tag in new_tags:
        print(f"Tagging image '{tag}' as: {new_tag}")
        docker_args = [
            "docker",
            "tag",
            tag,
            new_tag,
        ]
        try:
            run_docker_command(
                docker_args,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise ConnectorImageBuildError(
                error_text=e.stderr,
                build_args=docker_args,
            ) from e


def build_connector_image(
    connector_name: str,
    connector_directory: Path,
    metadata: MetadataFile,
    tag: str,
    primary_arch: ArchEnum = ArchEnum.ARM64,  # Assume MacBook M series by default
    no_verify: bool = False,
    dockerfile_override: Path | None = None,
) -> None:
    """Build a connector Docker image.

    This command builds a Docker image for a connector, using either
    the connector's Dockerfile or a base image specified in the metadata.
    The image is built for both AMD64 and ARM64 architectures.

    Args:
        connector_name: The name of the connector.
        connector_directory: The directory containing the connector code.
        metadata: The metadata of the connector.
        tag: The tag to apply to the built image.
        primary_arch: The primary architecture for the build (default: arm64). This
            architecture will be used for the same-named tag. Both AMD64 and ARM64
            images will be built, with the suffixes '-amd64' and '-arm64'.
        no_verify: If True, skip verification of the built image.

    Raises:
        ValueError: If the connector build options are not defined in metadata.yaml.
        ConnectorImageBuildError: If the image build or tag operation fails.
    """
    connector_kebab_name = connector_name

    if dockerfile_override:
        dockerfile_path = dockerfile_override
    else:
        dockerfile_path = connector_directory / "build" / "docker" / "Dockerfile"
        dockerignore_path = connector_directory / "build" / "docker" / "Dockerfile.dockerignore"
        try:
            dockerfile_text, dockerignore_text = get_dockerfile_templates(
                metadata=metadata,
                connector_directory=connector_directory,
            )
        except FileNotFoundError:
            # If the Dockerfile and .dockerignore are not found in the connector directory,
            # download the templates from the Airbyte repo. This is a fallback
            # in case the Airbyte repo not checked out locally.
            try:
                dockerfile_text, dockerignore_text = _download_dockerfile_defs(
                    connector_language=metadata.data.language,
                )
            except requests.HTTPError as e:
                raise ConnectorImageBuildError(
                    build_args=[],
                    error_text=(
                        "Could not locate local dockerfile templates and "
                        f"failed to download Dockerfile templates from github: {e}"
                    ),
                ) from e

        dockerfile_path.write_text(dockerfile_text)
        dockerignore_path.write_text(dockerignore_text)

    extra_build_script: str = ""
    build_customization_path = connector_directory / "build_customization.py"
    if build_customization_path.exists():
        extra_build_script = str(build_customization_path)

    dockerfile_path.parent.mkdir(parents=True, exist_ok=True)
    if not metadata.data.connectorBuildOptions:
        raise ValueError(
            "Connector build options are not defined in metadata.yaml. "
            "Please check the connector's metadata file."
        )

    base_image = metadata.data.connectorBuildOptions.baseImage
    build_args: dict[str, str | None] = {
        "BASE_IMAGE": base_image,
        "CONNECTOR_NAME": connector_kebab_name,
        "EXTRA_BUILD_SCRIPT": extra_build_script,
    }

    base_tag = f"{metadata.data.dockerRepository}:{tag}"
    arch_images: list[str] = []

    if metadata.data.language == ConnectorLanguage.JAVA:
        # This assumes that the repo root ('airbyte') is three levels above the
        # connector directory (airbyte/airbyte-integrations/connectors/source-foo).
        repo_root = connector_directory.parent.parent.parent
        # For Java connectors, we need to build the connector tar file first.
        subprocess.run(
            [
                "./gradlew",
                f":airbyte-integrations:connectors:{connector_name}:distTar",
            ],
            cwd=repo_root,
            text=True,
            check=True,
        )

    for arch in [ArchEnum.AMD64, ArchEnum.ARM64]:
        docker_tag = f"{base_tag}-{arch.value}"
        docker_tag_parts = docker_tag.split("/")
        if len(docker_tag_parts) > 2:
            docker_tag = "/".join(docker_tag_parts[-1:])
        arch_images.append(
            _build_image(
                context_dir=connector_directory,
                dockerfile=dockerfile_path,
                metadata=metadata,
                tag=docker_tag,
                arch=arch,
                build_args=build_args,
            )
        )

    _tag_image(
        tag=f"{base_tag}-{primary_arch.value}",
        new_tags=[base_tag],
    )
    if not no_verify:
        if verify_connector_image(base_tag):
            click.echo(f"Build completed successfully: {base_tag}")
            sys.exit(0)
        else:
            click.echo(f"Built image failed verification: {base_tag}", err=True)
            sys.exit(1)
    else:
        click.echo(f"Build completed successfully (without verification): {base_tag}")
        sys.exit(0)


def _download_dockerfile_defs(
    connector_language: ConnectorLanguage,
) -> tuple[str, str]:
    """Download the Dockerfile and .dockerignore templates for the specified connector language.

    We use the requests library to download from the master branch hosted on GitHub.

    Args:
        connector_language: The language of the connector.

    Returns:
        A tuple containing the Dockerfile and .dockerignore templates as strings.

    Raises:
        ValueError: If the connector language is not supported.
        requests.HTTPError: If the download fails.
    """
    print("Downloading Dockerfile and .dockerignore templates from GitHub...")
    # Map ConnectorLanguage to template directory
    language_to_template_suffix = {
        ConnectorLanguage.PYTHON: "python-connector",
        ConnectorLanguage.JAVA: "java-connector",
        ConnectorLanguage.MANIFEST_ONLY: "manifest-only-connector",
    }

    if connector_language not in language_to_template_suffix:
        raise ValueError(f"Unsupported connector language: {connector_language}")

    template_suffix = language_to_template_suffix[connector_language]
    base_url = f"https://github.com/airbytehq/airbyte/raw/master/docker-images/"

    dockerfile_url = f"{base_url}/Dockerfile.{template_suffix}"
    dockerignore_url = f"{base_url}/Dockerfile.{template_suffix}.dockerignore"

    dockerfile_resp = requests.get(dockerfile_url)
    dockerfile_resp.raise_for_status()
    dockerfile_text = dockerfile_resp.text

    dockerignore_resp = requests.get(dockerignore_url)
    dockerignore_resp.raise_for_status()
    dockerignore_text = dockerignore_resp.text

    return dockerfile_text, dockerignore_text


def get_dockerfile_templates(
    metadata: MetadataFile,
    connector_directory: Path,
) -> tuple[str, str]:
    """Get the Dockerfile template for the connector.

    Args:
        metadata: The metadata of the connector.
        connector_name: The name of the connector.

    Raises:
        ValueError: If the connector language is not supported.
        FileNotFoundError: If the Dockerfile or .dockerignore is not found.

    Returns:
        A tuple containing the Dockerfile and .dockerignore templates as strings.
    """
    if metadata.data.language not in [
        ConnectorLanguage.PYTHON,
        ConnectorLanguage.MANIFEST_ONLY,
        ConnectorLanguage.JAVA,
    ]:
        raise ValueError(
            f"Unsupported connector language: {metadata.data.language}. "
            "Please check the connector's metadata file."
        )

    airbyte_repo_root = resolve_airbyte_repo_root(
        from_dir=connector_directory,
    )
    # airbyte_repo_root successfully resolved
    dockerfile_path = (
        airbyte_repo_root / "docker-images" / f"Dockerfile.{metadata.data.language.value}-connector"
    )
    dockerignore_path = (
        airbyte_repo_root
        / "docker-images"
        / f"Dockerfile.{metadata.data.language.value}-connector.dockerignore"
    )
    if not dockerfile_path.exists():
        raise FileNotFoundError(
            f"Dockerfile for {metadata.data.language.value} connector not found at {dockerfile_path}"
        )
    if not dockerignore_path.exists():
        raise FileNotFoundError(
            f".dockerignore for {metadata.data.language.value} connector not found at {dockerignore_path}"
        )

    return dockerfile_path.read_text(), dockerignore_path.read_text()


def run_docker_command(
    cmd: list[str],
    *,
    check: bool = True,
    capture_output: bool = False,
) -> subprocess.CompletedProcess[str]:
    """Run a Docker command as a subprocess.

    Args:
        cmd: The command to run as a list of strings.
        check: If True, raises an exception if the command fails. If False, the caller is
            responsible for checking the return code.
        capture_output: If True, captures stdout and stderr and returns to the caller.
            If False, the output is printed to the console.

    Raises:
        subprocess.CalledProcessError: If the command fails and check is True.
    """
    print(f"Running command: {' '.join(cmd)}")

    process = subprocess.run(
        cmd,
        text=True,
        check=check,
        # If capture_output=True, stderr and stdout are captured and returned to caller:
        capture_output=capture_output,
        env={**os.environ, "DOCKER_BUILDKIT": "1"},
    )
    return process


def verify_docker_installation() -> bool:
    """Verify Docker is installed and running."""
    try:
        run_docker_command(["docker", "--version"])
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False


def verify_connector_image(
    image_name: str,
) -> bool:
    """Verify the built image by running the spec command.

    Args:
        image_name: The full image name with tag.

    Returns:
        True if the spec command succeeds, False otherwise.
    """
    logger.info(f"Verifying image {image_name} with 'spec' command...")

    cmd = ["docker", "run", "--rm", image_name, "spec"]

    try:
        result = run_docker_command(
            cmd,
            check=True,
            capture_output=True,
        )
        # check that the output is valid JSON
        if result.stdout:
            found_spec_output = False
            for line in result.stdout.split("\n"):
                if line.strip():
                    try:
                        # Check if the line is a valid JSON object
                        msg = json.loads(line)
                        if isinstance(msg, dict) and "type" in msg and msg["type"] == "SPEC":
                            found_spec_output = True

                    except json.JSONDecodeError as e:
                        logger.warning(f"Invalid JSON output from spec command: {e}: {line}")

            if not found_spec_output:
                logger.error("No valid JSON output found for spec command.")
                return False
        else:
            logger.error("No output from spec command.")
            return False
    except subprocess.CalledProcessError as e:
        logger.error(f"Image verification failed: {e.stderr}")
        return False

    return True



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.


from enum import Enum

_TERMINAL = True


class AsyncJobStatus(Enum):
    RUNNING = ("RUNNING", not _TERMINAL)
    COMPLETED = ("COMPLETED", _TERMINAL)
    FAILED = ("FAILED", _TERMINAL)
    TIMED_OUT = ("TIMED_OUT", _TERMINAL)

    def __init__(self, value: str, is_terminal: bool) -> None:
        self._value = value
        self._is_terminal = is_terminal

    def is_terminal(self) -> bool:
        """
        A status is terminal when a job status can't be updated anymore. For example if a job is completed, it will stay completed but a
        running job might because completed, failed or timed out.
        """
        return self._is_terminal



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
import logging
import uuid
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Any, Dict, Iterable, Mapping, Optional

import requests
from requests import Response

from airbyte_cdk import AirbyteMessage
from airbyte_cdk.logger import lazy_log
from airbyte_cdk.models import FailureType, Type
from airbyte_cdk.sources.declarative.async_job.job import AsyncJob
from airbyte_cdk.sources.declarative.async_job.repository import AsyncJobRepository
from airbyte_cdk.sources.declarative.async_job.status import AsyncJobStatus
from airbyte_cdk.sources.declarative.extractors.dpath_extractor import (
    DpathExtractor,
    RecordExtractor,
)
from airbyte_cdk.sources.declarative.extractors.response_to_file_extractor import (
    ResponseToFileExtractor,
)
from airbyte_cdk.sources.declarative.requesters.requester import Requester
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import SimpleRetriever
from airbyte_cdk.sources.http_logger import format_http_message
from airbyte_cdk.sources.types import Record, StreamSlice
from airbyte_cdk.utils import AirbyteTracedException

LOGGER = logging.getLogger("airbyte")


@dataclass
class AsyncHttpJobRepository(AsyncJobRepository):
    """
    See Readme file for more details about flow.
    """

    creation_requester: Requester
    polling_requester: Requester
    download_retriever: SimpleRetriever
    abort_requester: Optional[Requester]
    delete_requester: Optional[Requester]
    status_extractor: DpathExtractor
    status_mapping: Mapping[str, AsyncJobStatus]
    download_target_extractor: DpathExtractor

    # timeout for the job to be completed, passed from `polling_job_timeout`
    job_timeout: Optional[timedelta] = None

    record_extractor: RecordExtractor = field(
        init=False, repr=False, default_factory=lambda: ResponseToFileExtractor({})
    )
    download_target_requester: Optional[Requester] = (
        None  # use it in case polling_requester provides some <id> and extra request is needed to obtain list of urls to download from
    )

    def __post_init__(self) -> None:
        self._create_job_response_by_id: Dict[str, Response] = {}
        self._polling_job_response_by_id: Dict[str, Response] = {}

    def _get_validated_polling_response(self, stream_slice: StreamSlice) -> requests.Response:
        """
        Validates and retrieves the pooling response for a given stream slice.

        Args:
            stream_slice (StreamSlice): The stream slice to send the pooling request for.

        Returns:
            requests.Response: The validated pooling response.

        Raises:
            AirbyteTracedException: If the polling request returns an empty response.
        """

        polling_response: Optional[requests.Response] = self.polling_requester.send_request(
            stream_slice=stream_slice,
            log_formatter=lambda polling_response: format_http_message(
                response=polling_response,
                title="Async Job -- Polling",
                description="Poll the status of the server-side async job.",
                stream_name=None,
                is_auxiliary=True,
                type="ASYNC_POLL",
            ),
        )
        if polling_response is None:
            raise AirbyteTracedException(
                internal_message="Polling Requester received an empty Response.",
                failure_type=FailureType.system_error,
            )
        return polling_response

    def _get_validated_job_status(self, response: requests.Response) -> AsyncJobStatus:
        """
        Validates the job status extracted from the API response.

        Args:
            response (requests.Response): The API response.

        Returns:
            AsyncJobStatus: The validated job status.

        Raises:
            ValueError: If the API status is unknown.
        """

        api_status = next(iter(self.status_extractor.extract_records(response)), None)
        job_status = self.status_mapping.get(str(api_status), None)
        if job_status is None:
            raise ValueError(
                f"API status `{api_status}` is unknown. Contact the connector developer to make sure this status is supported."
            )

        return job_status

    def _start_job_and_validate_response(self, stream_slice: StreamSlice) -> requests.Response:
        """
        Starts a job and validates the response.

        Args:
            stream_slice (StreamSlice): The stream slice to be used for the job.

        Returns:
            requests.Response: The response from the job creation requester.

        Raises:
            AirbyteTracedException: If no response is received from the creation requester.
        """

        response: Optional[requests.Response] = self.creation_requester.send_request(
            stream_slice=stream_slice,
            log_formatter=lambda response: format_http_message(
                response=response,
                title="Async Job -- Create",
                description=f"Create the server-side async job. Timeout after: {self.job_timeout}",
                stream_name=None,
                is_auxiliary=True,
                type="ASYNC_CREATE",
            ),
        )

        if not response:
            raise AirbyteTracedException(
                internal_message="Always expect a response or an exception from creation_requester",
                failure_type=FailureType.system_error,
            )

        return response

    def start(self, stream_slice: StreamSlice) -> AsyncJob:
        """
        Starts a job for the given stream slice.

        Args:
            stream_slice (StreamSlice): The stream slice to start the job for.

        Returns:
            AsyncJob: The asynchronous job object representing the started job.
        """

        response: requests.Response = self._start_job_and_validate_response(stream_slice)
        job_id: str = str(uuid.uuid4())
        self._create_job_response_by_id[job_id] = response

        return AsyncJob(api_job_id=job_id, job_parameters=stream_slice, timeout=self.job_timeout)

    def update_jobs_status(self, jobs: Iterable[AsyncJob]) -> None:
        """
        Updates the status of multiple jobs.

        Because we don't have interpolation on random fields, we have this hack which consist on using the stream_slice to allow for
        interpolation. We are looking at enabling interpolation on more field which would require a change to those three layers:
        HttpRequester, RequestOptionProvider, RequestInputProvider.

        Args:
            jobs (Iterable[AsyncJob]): An iterable of AsyncJob objects representing the jobs to update.

        Returns:
            None
        """
        for job in jobs:
            stream_slice = self._get_create_job_stream_slice(job)
            polling_response: requests.Response = self._get_validated_polling_response(stream_slice)
            job_status: AsyncJobStatus = self._get_validated_job_status(polling_response)

            if job_status != job.status():
                lazy_log(
                    LOGGER,
                    logging.DEBUG,
                    lambda: f"Status of job {job.api_job_id()} changed from {job.status()} to {job_status}",
                )
            else:
                lazy_log(
                    LOGGER,
                    logging.DEBUG,
                    lambda: f"Status of job {job.api_job_id()} is still {job.status()}",
                )

            job.update_status(job_status)
            if job_status == AsyncJobStatus.COMPLETED:
                self._polling_job_response_by_id[job.api_job_id()] = polling_response

    def fetch_records(self, job: AsyncJob) -> Iterable[Mapping[str, Any]]:
        """
        Fetches records from the given job.

        Args:
            job (AsyncJob): The job to fetch records from.

        Yields:
            Iterable[Mapping[str, Any]]: A generator that yields records as dictionaries.

        """

        for target_url in self._get_download_targets(job):
            job_slice = job.job_parameters()
            stream_slice = StreamSlice(
                partition=job_slice.partition,
                cursor_slice=job_slice.cursor_slice,
                extra_fields={
                    **job_slice.extra_fields,
                    "download_target": target_url,
                },
            )
            for message in self.download_retriever.read_records({}, stream_slice):
                if isinstance(message, Record):
                    yield message.data
                elif isinstance(message, AirbyteMessage):
                    if message.type == Type.RECORD:
                        yield message.record.data  # type: ignore  # message.record won't be None here as the message is a record
                elif isinstance(message, (dict, Mapping)):
                    yield message
                else:
                    raise TypeError(f"Unknown type `{type(message)}` for message")

        yield from []

    def abort(self, job: AsyncJob) -> None:
        if not self.abort_requester:
            return

        abort_response = self.abort_requester.send_request(
            stream_slice=self._get_create_job_stream_slice(job),
            log_formatter=lambda abort_response: format_http_message(
                response=abort_response,
                title="Async Job -- Abort",
                description="Abort the running server-side async job.",
                stream_name=None,
                is_auxiliary=True,
                type="ASYNC_ABORT",
            ),
        )

    def delete(self, job: AsyncJob) -> None:
        if not self.delete_requester:
            return

        delete_job_reponse = self.delete_requester.send_request(
            stream_slice=self._get_create_job_stream_slice(job),
            log_formatter=lambda delete_job_reponse: format_http_message(
                response=delete_job_reponse,
                title="Async Job -- Delete",
                description="Delete the specified job from the list of Jobs.",
                stream_name=None,
                is_auxiliary=True,
                type="ASYNC_DELETE",
            ),
        )
        self._clean_up_job(job.api_job_id())

    def _clean_up_job(self, job_id: str) -> None:
        del self._create_job_response_by_id[job_id]
        del self._polling_job_response_by_id[job_id]

    def _get_creation_response_interpolation_context(self, job: AsyncJob) -> Dict[str, Any]:
        """
        Returns the interpolation context for the creation response.

        Args:
            job (AsyncJob): The job for which to get the creation response interpolation context.

        Returns:
            Dict[str, Any]: The interpolation context as a dictionary.
        """
        # TODO: currently we support only JsonDecoder to decode the response to track the ids or the status
        # of the Jobs. We should consider to add the support of other decoders like XMLDecoder, in the future
        creation_response_context = dict(self._create_job_response_by_id[job.api_job_id()].json())
        if not "headers" in creation_response_context:
            creation_response_context["headers"] = self._create_job_response_by_id[
                job.api_job_id()
            ].headers
        if not "request" in creation_response_context:
            creation_response_context["request"] = self._create_job_response_by_id[
                job.api_job_id()
            ].request
        return creation_response_context

    def _get_polling_response_interpolation_context(self, job: AsyncJob) -> Dict[str, Any]:
        """
        Returns the interpolation context for the polling response.

        Args:
            job (AsyncJob): The job for which to get the polling response interpolation context.

        Returns:
            Dict[str, Any]: The interpolation context as a dictionary.
        """
        # TODO: currently we support only JsonDecoder to decode the response to track the ids or the status
        # of the Jobs. We should consider to add the support of other decoders like XMLDecoder, in the future
        polling_response_context = dict(self._polling_job_response_by_id[job.api_job_id()].json())
        if not "headers" in polling_response_context:
            polling_response_context["headers"] = self._polling_job_response_by_id[
                job.api_job_id()
            ].headers
        if not "request" in polling_response_context:
            polling_response_context["request"] = self._polling_job_response_by_id[
                job.api_job_id()
            ].request
        return polling_response_context

    def _get_create_job_stream_slice(self, job: AsyncJob) -> StreamSlice:
        return StreamSlice(
            partition=job.job_parameters().partition,
            cursor_slice=job.job_parameters().cursor_slice,
            extra_fields=dict(job.job_parameters().extra_fields)
            | {
                "creation_response": self._get_creation_response_interpolation_context(job),
            },
        )

    def _get_download_targets(self, job: AsyncJob) -> Iterable[str]:
        if not self.download_target_requester:
            url_response = self._polling_job_response_by_id[job.api_job_id()]
        else:
            stream_slice: StreamSlice = StreamSlice(
                partition={},
                cursor_slice={},
                extra_fields={
                    "polling_response": self._get_polling_response_interpolation_context(job),
                },
            )
            url_response = self.download_target_requester.send_request(stream_slice=stream_slice)  # type: ignore # we expect download_target_requester to always be presented, otherwise raise an exception as we cannot proceed with the report
            if not url_response:
                raise AirbyteTracedException(
                    internal_message="Always expect a response or an exception from download_target_requester",
                    failure_type=FailureType.system_error,
                )

        yield from self.download_target_extractor.extract_records(url_response)  # type: ignore # we expect download_target_extractor to always return list of strings



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Pytest hooks for Airbyte CDK tests.

These hooks are used to customize the behavior of pytest during test discovery and execution.

To use these hooks within a connector, add the following lines to the connector's `conftest.py`
file, or to another file that is imported during test discovery:

```python
pytest_plugins = [
    "airbyte_cdk.test.standard_tests.pytest_hooks",
]
```
"""

import pytest


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """
    A helper for pytest_generate_tests hook.

    If a test method (in a class subclassed from our base class)
    declares an argument 'scenario', this function retrieves the
    'scenarios' attribute from the test class and parametrizes that
    test with the values from 'scenarios'.

    ## Usage

    ```python
    from airbyte_cdk.test.standard_tests.connector_base import (
        generate_tests,
        ConnectorTestSuiteBase,
    )

    def pytest_generate_tests(metafunc):
        generate_tests(metafunc)

    class TestMyConnector(ConnectorTestSuiteBase):
        ...

    ```
    """
    # Check if the test function requires an 'scenario' argument
    if "scenario" in metafunc.fixturenames:
        # Retrieve the test class
        test_class = metafunc.cls
        if test_class is None:
            return

        # Get the 'scenarios' attribute from the class
        scenarios_attr = getattr(test_class, "get_scenarios", None)
        if scenarios_attr is None:
            raise ValueError(
                f"Test class {test_class} does not have a 'scenarios' attribute. "
                "Please define the 'scenarios' attribute in the test class."
            )

        scenarios = test_class.get_scenarios()
        ids = [str(scenario) for scenario in scenarios]
        metafunc.parametrize("scenario", scenarios, ids=ids)



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from dataclasses import dataclass

from airbyte_cdk.sources.declarative.types import Record

from .default_file_uploader import DefaultFileUploader
from .file_uploader import FileUploader


@dataclass
class ConnectorBuilderFileUploader(FileUploader):
    """
    Connector builder file uploader
    Acts as a decorator or wrapper around a FileUploader instance, copying the attributes from record.file_reference into the record.data.
    """

    file_uploader: DefaultFileUploader

    def upload(self, record: Record) -> None:
        self.file_uploader.upload(record=record)
        for file_reference_key, file_reference_value in record.file_reference.__dict__.items():
            if not file_reference_key.startswith("_"):
                record.data[file_reference_key] = file_reference_value  # type: ignore



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Optional

from airbyte_cdk.sources.file_based import AbstractFileBasedSpec
from airbyte_cdk.sources.file_based.remote_file import RemoteFile


class AbstractFileBasedStreamPermissionsReader(ABC):
    """
    This class is responsible for reading file permissions and Identities from a source.
    """

    def __init__(self) -> None:
        self._config = None

    @property
    def config(self) -> Optional[AbstractFileBasedSpec]:
        return self._config

    @config.setter
    @abstractmethod
    def config(self, value: AbstractFileBasedSpec) -> None:
        """
        FileBasedSource reads the config from disk and parses it, and once parsed, the source sets the config on its StreamReader.

        Note: FileBasedSource only requires the keys defined in the abstract config, whereas concrete implementations of StreamReader
        will require keys that (for example) allow it to authenticate with the 3rd party.

        Therefore, concrete implementations of AbstractFileBasedStreamPermissionsReader's's config setter should assert that `value` is of the correct
        config type for that type of StreamReader.
        """
        ...

    @abstractmethod
    def get_file_acl_permissions(self, file: RemoteFile, logger: logging.Logger) -> Dict[str, Any]:
        """
        This function should return the allow list for a given file, i.e. the list of all identities and their permission levels associated with it

        e.g.
        def get_file_acl_permissions(self, file: RemoteFile, logger: logging.Logger):
            api_conn = some_api.conn(credentials=SOME_CREDENTIALS)
            result = api_conn.get_file_permissions_info(file.id)
            return MyPermissionsModel(
                id=result["id"],
                access_control_list = result["access_control_list"],
                is_public = result["is_public"],
                ).dict()
        """
        ...

    @abstractmethod
    def load_identity_groups(self, logger: logging.Logger) -> Iterable[Dict[str, Any]]:
        """
        This function should return the Identities in a determined "space" or "domain" where the file metadata (ACLs) are fetched and ACLs items (Identities) exists.

        e.g.
        def load_identity_groups(self, logger: logging.Logger) -> Iterable[Dict[str, Any]]:
            api_conn = some_api.conn(credentials=SOME_CREDENTIALS)
            users_api = api_conn.users()
            groups_api = api_conn.groups()
            members_api = self.google_directory_service.members()
            for user in users_api.list():
                yield my_identity_model(id=user.id, name=user.name, email_address=user.email, type="user").dict()
            for group in groups_api.list():
                group_obj = my_identity_model(id=group.id, name=groups.name, email_address=user.email, type="group").dict()
                for member in members_api.list(group=group):
                    group_obj.member_email_addresses = group_obj.member_email_addresses or []
                    group_obj.member_email_addresses.append(member.email)
                yield group_obj.dict()
        """
        ...

    @property
    @abstractmethod
    def file_permissions_schema(self) -> Dict[str, Any]:
        """
        This function should return the permissions schema for file permissions stream.

        e.g.
        def file_permissions_schema(self) -> Dict[str, Any]:
            # you can also follow the pattern we have for python connectors and have a json file and read from there e.g. schemas/identities.json
            return {
                  "type": "object",
                  "properties": {
                    "id": { "type": "string" },
                    "file_path": { "type": "string" },
                    "access_control_list": {
                      "type": "array",
                      "items": { "type": "string" }
                    },
                    "publicly_accessible": { "type": "boolean" }
                  }
                }
        """
        ...

    @property
    @abstractmethod
    def identities_schema(self) -> Dict[str, Any]:
        """
        This function should return the identities schema for file identity stream.

        e.g.
        def identities_schema(self) -> Dict[str, Any]:
            # you can also follow the pattern we have for python connectors and have a json file and read from there e.g. schemas/identities.json
            return {
              "type": "object",
              "properties": {
                "id": { "type": "string" },
                "remote_id": { "type": "string" },
                "name": { "type": ["null", "string"] },
                "email_address": { "type": ["null", "string"] },
                "member_email_addresses": { "type": ["null", "array"] },
                "type": { "type": "string" },
              }
            }
        """
        ...



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import csv
import gzip
import io
import json
import logging
from dataclasses import dataclass
from io import BufferedIOBase, TextIOWrapper
from typing import Any, Optional

import orjson
import requests

from airbyte_cdk.models import FailureType
from airbyte_cdk.sources.declarative.decoders.decoder import DECODER_OUTPUT_TYPE, Decoder
from airbyte_cdk.sources.declarative.decoders.decoder_parser import (
    PARSER_OUTPUT_TYPE,
    PARSERS_BY_HEADER_TYPE,
    PARSERS_TYPE,
    Parser,
)
from airbyte_cdk.utils import AirbyteTracedException

logger = logging.getLogger("airbyte")


@dataclass
class GzipParser(Parser):
    inner_parser: Parser

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        """
        Decompress gzipped bytes and pass decompressed data to the inner parser.

        IMPORTANT:
            - If the data is not gzipped, reset the pointer and pass the data to the inner parser as is.

        Note:
            - The data is not decoded by default.
        """

        with gzip.GzipFile(fileobj=data, mode="rb") as gzipobj:
            yield from self.inner_parser.parse(gzipobj)


@dataclass
class JsonParser(Parser):
    encoding: str = "utf-8"

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        """
        Attempts to deserialize data using orjson library. As an extra layer of safety we fallback on the json library to deserialize the data.
        """
        raw_data = data.read()
        body_json = self._parse_orjson(raw_data) or self._parse_json(raw_data)

        if body_json is None:
            raise AirbyteTracedException(
                message="Response JSON data failed to be parsed. See logs for more information.",
                internal_message=f"Response JSON data failed to be parsed.",
                failure_type=FailureType.system_error,
            )

        if isinstance(body_json, list):
            yield from body_json
        else:
            yield from [body_json]

    def _parse_orjson(self, raw_data: bytes) -> Optional[Any]:
        try:
            return orjson.loads(raw_data.decode(self.encoding))
        except Exception as exc:
            logger.debug(
                f"Failed to parse JSON data using orjson library. Falling back to json library. {exc}"
            )
            return None

    def _parse_json(self, raw_data: bytes) -> Optional[Any]:
        try:
            return json.loads(raw_data.decode(self.encoding))
        except Exception as exc:
            logger.error(f"Failed to parse JSON data using json library. {exc}")
            return None


@dataclass
class JsonLineParser(Parser):
    encoding: Optional[str] = "utf-8"

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        for line in data:
            try:
                yield json.loads(line.decode(encoding=self.encoding or "utf-8"))
            except json.JSONDecodeError as e:
                logger.warning(f"Cannot decode/parse line {line!r} as JSON, error: {e}")


@dataclass
class CsvParser(Parser):
    # TODO: migrate implementation to re-use file-base classes
    encoding: Optional[str] = "utf-8"
    delimiter: Optional[str] = ","

    def _get_delimiter(self) -> Optional[str]:
        """
        Get delimiter from the configuration. Check for the escape character and decode it.
        """
        if self.delimiter is not None:
            if self.delimiter.startswith("\\"):
                self.delimiter = self.delimiter.encode("utf-8").decode("unicode_escape")

        return self.delimiter

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        """
        Parse CSV data from decompressed bytes.
        """
        text_data = TextIOWrapper(data, encoding=self.encoding)  # type: ignore
        reader = csv.DictReader(text_data, delimiter=self._get_delimiter() or ",")
        for row in reader:
            yield row


class CompositeRawDecoder(Decoder):
    """
    Decoder strategy to transform a requests.Response into a PARSER_OUTPUT_TYPE
    passed response.raw to parser(s).

    Note: response.raw is not decoded/decompressed by default. Parsers should be instantiated recursively.

    Example:
        composite_raw_decoder = CompositeRawDecoder(
            parser=GzipParser(
                inner_parser=JsonLineParser(encoding="iso-8859-1")
            )
        )
    """

    def __init__(
        self,
        parser: Parser,
        stream_response: bool = True,
        parsers_by_header: PARSERS_BY_HEADER_TYPE = None,
    ) -> None:
        # since we moved from using `dataclass` to `__init__` method,
        # we need to keep using the `parser` to be able to resolve the depenencies
        # between the parsers correctly.
        self.parser = parser

        self._parsers_by_header = parsers_by_header if parsers_by_header else {}
        self._stream_response = stream_response

    @classmethod
    def by_headers(
        cls,
        parsers: PARSERS_TYPE,
        stream_response: bool,
        fallback_parser: Parser,
    ) -> "CompositeRawDecoder":
        """
        Create a CompositeRawDecoder instance based on header values.

        Args:
            parsers (PARSERS_TYPE): A list of tuples where each tuple contains headers, header values, and a parser.
            stream_response (bool): A flag indicating whether the response should be streamed.
            fallback_parser (Parser): A parser to use if no matching header is found.

        Returns:
            CompositeRawDecoder: An instance of CompositeRawDecoder configured with the provided parsers.
        """
        parsers_by_header = {}
        for headers, header_values, parser in parsers:
            for header in headers:
                parsers_by_header[header] = {header_value: parser for header_value in header_values}
        return cls(fallback_parser, stream_response, parsers_by_header)

    def is_stream_response(self) -> bool:
        return self._stream_response

    def decode(self, response: requests.Response) -> DECODER_OUTPUT_TYPE:
        parser = self._select_parser(response)
        if self.is_stream_response():
            # urllib mentions that some interfaces don't play nice with auto_close
            # More info here: https://urllib3.readthedocs.io/en/stable/user-guide.html#using-io-wrappers-with-response-content
            # We have indeed observed some issues with CSV parsing.
            # Hence, we will manage the closing of the file ourselves until we find a better solution.
            response.raw.auto_close = False
            yield from parser.parse(
                data=response.raw,  # type: ignore[arg-type]
            )
            response.raw.close()
        else:
            yield from parser.parse(data=io.BytesIO(response.content))

    def _select_parser(self, response: requests.Response) -> Parser:
        """
        Selects the appropriate parser based on the response headers.

        This method iterates through the `_parsers_by_header` dictionary to find a matching parser
        based on the headers in the response. If a matching header and header value are found,
        the corresponding parser is returned. If no match is found, the default parser is returned.

        Args:
            response (requests.Response): The HTTP response object containing headers to check.

        Returns:
            Parser: The parser corresponding to the matched header value, or the default parser if no match is found.
        """
        for header, parser_by_header_value in self._parsers_by_header.items():
            if (
                header in response.headers
                and response.headers[header] in parser_by_header_value.keys()
            ):
                return parser_by_header_value[response.headers[header]]
        return self.parser



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#


from abc import ABC, abstractmethod
from typing import Any, Optional

from airbyte_cdk.sources.types import Config


class Interpolation(ABC):
    """
    Strategy for evaluating the interpolated value of a string at runtime using Jinja.
    """

    @abstractmethod
    def eval(
        self,
        input_str: str,
        config: Config,
        default: Optional[str] = None,
        **additional_options: Any,
    ) -> Any:
        """
        Interpolates the input string using the config, and additional options passed as parameter.

        :param input_str: The string to interpolate
        :param config: The user-provided configuration as specified by the source's spec
        :param default: Default value to return if the evaluation returns an empty string
        :param additional_options: Optional parameters used for interpolation
        :return: The interpolated string
        """
        pass



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Mapping, Optional, Union

from airbyte_cdk.sources.declarative.interpolation import InterpolatedString
from airbyte_cdk.sources.types import Config


@dataclass
class ConcurrencyLevel:
    """
    Returns the number of worker threads that should be used when syncing concurrent streams in parallel

    Attributes:
        default_concurrency (Union[int, str]): The hardcoded integer or interpolation of how many worker threads to use during a sync
        max_concurrency (Optional[int]): The maximum number of worker threads to use when the default_concurrency is exceeded
    """

    default_concurrency: Union[int, str]
    max_concurrency: Optional[int]
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if isinstance(self.default_concurrency, int):
            self._default_concurrency: Union[int, InterpolatedString] = self.default_concurrency
        elif "config" in self.default_concurrency and not self.max_concurrency:
            raise ValueError(
                "ConcurrencyLevel requires that max_concurrency be defined if the default_concurrency can be used-specified"
            )
        else:
            self._default_concurrency = InterpolatedString.create(
                self.default_concurrency, parameters=parameters
            )

    def get_concurrency_level(self) -> int:
        if isinstance(self._default_concurrency, InterpolatedString):
            evaluated_default_concurrency = self._default_concurrency.eval(config=self.config)
            if not isinstance(evaluated_default_concurrency, int):
                raise ValueError("default_concurrency did not evaluate to an integer")
            return (
                min(evaluated_default_concurrency, self.max_concurrency)
                if self.max_concurrency
                else evaluated_default_concurrency
            )
        else:
            return self._default_concurrency



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from typing import Any, Mapping

from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.migrations.state_migration import StateMigration
from airbyte_cdk.sources.declarative.models import (
    CustomIncrementalSync,
    DatetimeBasedCursor,
    SubstreamPartitionRouter,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import ParentStreamConfig


def _is_already_migrated(stream_state: Mapping[str, Any]) -> bool:
    return "states" in stream_state


class LegacyToPerPartitionStateMigration(StateMigration):
    """
    Transforms the input state for per-partitioned streams from the legacy format to the low-code format.
    The cursor field and partition ID fields are automatically extracted from the stream's DatetimebasedCursor and SubstreamPartitionRouter.

    Example input state:
    {
    "13506132": {
      "last_changed": "2022-12-27T08:34:39+00:00"
    }
    Example output state:
    {
      "partition": {"id": "13506132"},
      "cursor": {"last_changed": "2022-12-27T08:34:39+00:00"}
    }
    """

    def __init__(
        self,
        partition_router: SubstreamPartitionRouter,
        cursor: CustomIncrementalSync | DatetimeBasedCursor,
        config: Mapping[str, Any],
        parameters: Mapping[str, Any],
    ):
        self._partition_router = partition_router
        self._cursor = cursor
        self._config = config
        self._parameters = parameters
        self._partition_key_field = InterpolatedString.create(
            self._get_partition_field(self._partition_router), parameters=self._parameters
        ).eval(self._config)
        self._cursor_field = InterpolatedString.create(
            self._cursor.cursor_field, parameters=self._parameters
        ).eval(self._config)

    def _get_partition_field(self, partition_router: SubstreamPartitionRouter) -> str:
        parent_stream_config = partition_router.parent_stream_configs[0]

        # Retrieve the partition field with a condition, as properties are returned as a dictionary for custom components.
        partition_field = (
            parent_stream_config.partition_field
            if isinstance(parent_stream_config, ParentStreamConfig)
            else parent_stream_config.get("partition_field")  # type: ignore # See above comment on why parent_stream_config might be a dict
        )

        return partition_field

    def should_migrate(self, stream_state: Mapping[str, Any]) -> bool:
        if _is_already_migrated(stream_state):
            return False

        # There is exactly one parent stream
        number_of_parent_streams = len(self._partition_router.parent_stream_configs)  # type: ignore # custom partition will introduce this attribute if needed
        if number_of_parent_streams != 1:
            # There should be exactly one parent stream
            return False
        """
        The expected state format is
        "<parent_key_id>" : {
          "<cursor_field>" : "<cursor_value>"
        }
        """
        if stream_state:
            for key, value in stream_state.items():
                if isinstance(value, dict):
                    keys = list(value.keys())
                    if len(keys) != 1:
                        # The input partitioned state should only have one key
                        return False
                    if keys[0] != self._cursor_field:
                        # Unexpected key. Found {keys[0]}. Expected {self._cursor.cursor_field}
                        return False
        return True

    def migrate(self, stream_state: Mapping[str, Any]) -> Mapping[str, Any]:
        states = [
            {"partition": {self._partition_key_field: key}, "cursor": value}
            for key, value in stream_state.items()
        ]
        return {"states": states}
